require 'common.util'


GLHandle = class()
function GLHandle:delete()
	gl.glDeleteObject(self.id)
end
function GLHandle:getInfoLog()
	return gl.glGetInfoLog(self.id)		-- should this return nil or '' on no characters?
end


Shader = class(GLHandle)
function Shader:init(args)
	if not args then args = {} end
	if type(args) == 'string' then args = {source=args} end
	self.id = gl.glCreateShaderObject(args.type or self.type)
	if args.source then
		self:setSource(args.source)
		if not args.deferCompile then
			self:compile()
		end
	end
end
function Shader:setSource(buffer)
	gl.glShaderSource(self.id, buffer)
end
function Shader:compile()
	gl.glCompileShader(self.id)
end
function Shader:getCompileStatus()
	return gl.glGetObjectParameter(self.id, gl.GL_OBJECT_COMPILE_STATUS)
end


VertexShader = class(Shader)
VertexShader.type = gl.GL_VERTEX_SHADER


FragmentShader = class(Shader)
FragmentShader.type = gl.GL_FRAGMENT_SHADER


ShaderProgram = class(GLHandle)
function ShaderProgram:init(args)
	self.id = gl.glCreateProgramObject()
	-- TODO - like my C++ alternative, should this be capable of keeping track of attached objects for cascading 'delete' call?
	local needsLink
	if args.vertexCode then
		local vs = VertexShader.new(args.vertexCode)
		--TODO - automatically print errors / assert failures on compile?
		-- if we just returned status / message then VertexShader.new() would need a parameter for asserting compiling wokred
		-- ... or to return the status through the new function ...
		local status = vs:getCompileStatus()
		if status ~= 1 then
			error('vertex shader:\n'..tostring(vs:getInfoLog())..'\ncompile failed with status '..tostring(status))
		end
		self:attach(vs)
		needsLink = true
	end
	if args.fragmentCode then
		local fs = FragmentShader.new(args.fragmentCode)
		local status = fs:getCompileStatus()
		if status ~= 1 then
			error('fragment shader:\n'..tostring(fs:getInfoLog())..'\ncompile failed with status '..tostring(status))
		end
		self:attach(fs)
		needsLink = true
	end
	if args.attach then
		for _,v in ipairs(args.attach) do
			self:attach(v)
		end
		needsLink = true
	end
	if needsLink and not args.deferLink then
		self:link()
	end
	if args.uniforms then
		self:use()
		for k,v in pairs(args.uniforms) do
			gl.glUniform(k, v)
		end
	end
end
function ShaderProgram:attach(handle)
	gl.glAttachObject(self.id, handle.id)
end
function ShaderProgram:link()
	gl.glLinkProgram(self.id)
end
function ShaderProgram:getLinkStatus()
	return gl.glGetObjectParameter(self.id, gl.GL_OBJECT_LINK_STATUS)
end
function ShaderProgram:use()
	gl.glUseProgramObject(self.id)
end
function ShaderProgram.useNone()
	gl.glUseProgramObject(0)
end
function ShaderProgram:getUniformLocation(name)
	return gl.glGetUniformLocation(self.id, name)
end
function ShaderProgram:setUniform(name, ...)
	self:use()
	gl.glUniform(name, ...)
end
function ShaderProgram:setUniformf(name, ...)
	self:use()
	gl.glUniformf(name, ...)
end
function ShaderProgram:setUniformi(name, ...)
	self:use()
	gl.glUniformi(name, ...)
end


KernelShader = class(ShaderProgram)
function KernelShader:init(args)
	local varyingCode = [[
varying vec2 pos;
]]
	local vertexCode = varyingCode..[[
void main() {
	pos = gl_Vertex.xy;
	gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
}
]]
	local fragmentCodePrefix = varyingCode;
	local uniforms = table.new()
	if args.uniforms then
		for uniformName,uniformType in pairs(args.uniforms) do
			if type(uniformType) == 'table' then	-- if we have a table then assume it's a type/value pair
				uniforms[uniformName] = uniformType[2]
				uniformType = uniformType[1]
			end
			fragmentCodePrefix = fragmentCodePrefix .. 'uniform '..uniformType..' '..uniformName..';\n'
		end
	end
	if args.texs then
		for i,n in ipairs(args.texs) do
			uniforms[n] = i-1
			fragmentCodePrefix = fragmentCodePrefix .. 'uniform sampler2D '..n..';\n'
		end
	end
	KernelShader.super.init(self, {
		vertexCode=args.vertexCode or vertexCode;
		fragmentCode=fragmentCodePrefix..assert(args.code);
		uniforms=uniforms;
	})
end

